(** The universe of our set (theory). *)
Parameter S : Type.

(** We need some Relation "is element of" *)
Parameter elem : S -> S -> Prop.

(** Axiom of extensionality *)
Axiom extensionality : 
  forall x y,
  x = y <-> 
  forall z, 
    elem z x <-> 
    elem z y.

(** Existence of an empty set *)
Parameter empty_set : S.

Definition empty_property s :=
  forall x,
    ~ elem x s.

Axiom empty : 
  empty_property empty_set.

(** Thanks to the [empty_set], our domain [S] cannot be empty. *)
Fact domain_nonempty : inhabited S.
Proof.
  constructor;
  exact empty_set.
Qed.

(** Axiom of regularity/foundation *)
Axiom regularity : 
  forall a, 
    a <> empty_set ->
    exists b, 
      elem b a /\
      ~ exists c, 
        elem c a /\ 
        elem c b.

Lemma not_empty_contains_element :
  forall a,
    a <> empty_set ->
    exists b,
      elem b a.
Proof.
  intros a H1.
  apply regularity in H1 as [b [H1 ?]].
  firstorder.
Qed.

(** Axiom of union *)
Parameter union_set : S -> S.

Definition union_property a s :=
  forall c, 
    elem c s <-> 
    exists d, 
      elem d a /\ 
      elem c d.

Axiom union : 
  forall a,
    union_property a (union_set a).

Lemma union_empty_set : 
  union_set empty_set = empty_set.
Proof.
  apply extensionality.
  intros z;
  split;
  intros H1;
  [apply union in H1 as [d [H1 H2]] | apply union]; 
  apply empty in H1; contradiction.
Qed.

(** Axiom schema of replacement *)
Axiom replacement : 
  forall (F : S -> S -> Prop), 
    (
      forall x y z,
        F x y -> 
        F x z -> 
        y = z
    ) -> 
    forall a, 
      { b : S | 
        forall c, 
          elem c b <->
          exists d, 
            elem d a /\ 
            F d c}.

(** "Axiom" of specification and the existence of subsets. *)

Definition sub_set_sig (P : S -> Prop) (a : S) : {b : S | forall c, elem c b <-> exists d, elem d a /\ P d /\ d = c}.
Proof.
  apply replacement.
  intros x y z [H1 H2] [H3 H4]; subst; reflexivity.
Defined.

Definition sub_set P a := proj1_sig (sub_set_sig P a).

Definition specification_property P a s :=
  forall c, 
    elem c s <-> 
    (
      elem c a /\ 
      P c
    ).

Theorem specification : 
  forall P a, 
    specification_property P a (sub_set P a).
Proof.
  intros P a c; 
  split; 
  intros H1; 
  pose proof (proj2_sig (sub_set_sig P a)) as H2; 
  simpl in H2;
  destruct (H2 c) as [H3 ?]; 
  [destruct (H3 H1) | destruct H1]; 
      firstorder; 
      subst; 
      firstorder.
Qed.

Definition intersection_set a (not_empty : a <> empty_set) := sub_set (fun b => forall c, elem c a -> elem b c) (union_set a).

Definition intersection_property a s :=
  forall c, 
    elem c s <-> 
    forall d, 
      elem d a ->
      elem c d.

Theorem intersection : 
  forall a (not_empty : a <> empty_set),
  intersection_property a (intersection_set a not_empty).
Proof.
  intros a not_empty b;
  split;
  unfold intersection_set;
  intros H1.
  -
    intros d H4.
    apply specification in H1 as [H1 H3].
    apply union in H1.
    firstorder.
  -
    apply specification.
    split.
    +
      apply union.
      apply not_empty_contains_element in not_empty as [c H2].
      firstorder.
    +
      assumption.
Qed.

Definition diff a b := sub_set (fun x => ~ elem x b) a.

(** Axiom of power set *)
Parameter pot_set : S -> S.

Definition pot_property a s :=
  forall b, 
    elem b s <-> 
    forall c, 
      elem c b ->
      elem c a.

Axiom pot : 
  forall a, 
    pot_property a (pot_set a).

(** "Axiom" of pairing *)

Lemma pot_set_neq_empty_set : empty_set <> pot_set empty_set.
Proof.
  intros H1.
  pose proof (pot empty_set) as H2.
  pose proof empty as H3.
  rewrite H1 in H3.
  specialize (H3 empty_set).
  apply H3;
  apply H2;
  intros c H4;
  exact H4.
Qed.

Definition pair_set_sig (a b : S) : 
  {p : S | 
      forall c, 
        elem c p <-> 
        exists d, 
          let special_pair := pot_set (pot_set empty_set) in
          elem d special_pair /\
          (
            (d = empty_set /\ c = a) \/ 
            (d = pot_set empty_set /\ c = b))
  }.
Proof.
  apply replacement.
  intros x y z H1 H2.
  destruct H1 as [H3|H3], H2 as [H4|H4]; 
  destruct H3 as [H5 H6], H4 as [H7 H8]; 
  subst; 
  firstorder; 
  exfalso; 
  try apply pot_set_neq_empty_set; 
  easy.
Defined.

Definition pair_set a b := proj1_sig (pair_set_sig a b).

Definition pair_property a b s := 
  forall d,
    elem d s <->
    (
      d = a \/
      d = b
    ).

Theorem pair : 
  forall a b,
  pair_property a b (pair_set a b).
Proof.
  intros a b d; split; intros H1;
    pose proof (proj2_sig (pair_set_sig a b)) as H2;
    destruct (H2 d) as [H3 H4]; clear H2.
  -
    specialize (H3 H1).
    destruct H3 as [e H5].
    destruct H5; firstorder.
  -
    apply H4; clear H4 H3.
    destruct H1; subst;
    [exists empty_set | exists (pot_set empty_set)];
        split.
    +
      apply pot; intros; exfalso;
      eapply empty; eassumption.
    +
      left; split; reflexivity.
    +
      apply pot; intros; assumption.
    +
      right; split; reflexivity.
Qed.

Lemma pair_set_not_empty :
  forall a b,
    pair_set a b <> empty_set.
Proof.
  intros a b H1.
  eapply extensionality in H1.
  eapply empty.
  apply H1.
  apply pair.
  firstorder.
Qed.

Theorem pair_uniqueness : 
  forall a b c, 
    pair_property a b c -> 
    c = pair_set a b.
Proof.
  intros a b c H1;
  red in H1.
  apply extensionality;
  intros z; split; intros H2.
  -
    destruct (H1 z) as [H3 H4].
    destruct (H3 H2) as [H5|H5]; subst; apply pair; [left|right]; reflexivity.
  -
    destruct (pair a b z) as [H3 H4];
    destruct (H3 H2) as [H5|H5]; subst; apply H1; [left|right]; reflexivity.
Qed.

Theorem pair_set_commutative : 
  forall a b, 
    pair_set a b = pair_set b a.
Proof.
  intros;
  apply extensionality;
  intro;
  split;
  intros H1;
  apply pair in H1 as [|];
  apply pair;
  firstorder.
Qed.

Definition single_set a := pair_set a a.

Definition union2_set a b := union_set (pair_set a b).

Definition union2_property a b s :=
  forall c, 
    elem c s <-> 
    elem c a \/ elem c b.

Theorem union2 : 
  forall a b,
    union2_property a b (union2_set a b).
Proof.
  intros a b c;
  split; 
  intros H1.
  -
    apply union in H1 as [d [H1 H3]];
    apply pair in H1 as [H1|H1]; subst;
    firstorder.
  -
    apply union.
    destruct H1 as [H1|H1];
    eexists;
    split.
    +
      apply pair;
      firstorder.
    +
      assumption.
    +
      apply pair;
      right; 
      reflexivity.
    +
      assumption.
Qed.

Theorem union2_set_commutative : 
  forall a b,
    union2_set a b = union2_set b a.
Proof.
  intros;
  apply extensionality;
  intro;
  split;
  intro;
  unfold union2_set;
  rewrite pair_set_commutative;
  assumption.
Qed.

Theorem union2_set_associative :
  forall a b c,
    union2_set a (union2_set b c) = union2_set (union2_set a b) c.
Proof.
  intros a b c;
  apply extensionality;
  intro d;
  split;
  intros H1;
  apply union2 in H1 as [H1|H1];
  apply union2.
  -
    left; apply union2; firstorder.
  -
    apply union2 in H1 as [H1|H1].
    +
      left; apply union2; firstorder.
    +
      right; assumption.
  -
    apply union2 in H1 as [H1|H1].
    +
      left; assumption.
    +
      right; apply union2; firstorder.
  -
    right; apply union2; firstorder.
Qed.

Definition intersection2_set a b := intersection_set (pair_set a b) (pair_set_not_empty a b).

Definition intersection2_property a b s :=
  forall c, 
    elem c s <-> 
    elem c a /\ elem c b.

Theorem intersection2 : 
  forall a b,
  intersection2_property a b (intersection2_set a b).
Proof.
  unfold intersection2_set;
  intros a b c;
  split;
  intros H1.
  -
    assert (H2 : forall d, elem d (pair_set a b) -> elem c d).
    {
      eapply intersection; exact H1.  
    }
    split; apply H2; apply pair; firstorder.
  -
    destruct H1 as [H1 H2].
    apply intersection; intros d H3;
    apply pair in H3 as [|]; subst; firstorder.
Qed.

Theorem intersection2_set_commutative :
  forall a b,
    intersection2_set a b = intersection2_set b a.
Proof.
  intros a b;
  apply extensionality;
  intros c;
  split;
  intros H1;
  apply intersection2 in H1;
  apply intersection2;
  firstorder.
Qed.

Theorem intersection2_set_associative :
  forall a b c,
    intersection2_set a (intersection2_set b c) = intersection2_set (intersection2_set a b) c.
Proof.
  intros a b c;
  apply extensionality;
  intros d;
  split;
  intros H1;
  apply intersection2 in H1;
  apply intersection2;
  destruct H1 as [H1 H2];
  [apply intersection2 in H2 as [] 
  |apply intersection2 in H1 as []
  ];
  split;
  try apply intersection2; 
  firstorder.
Qed.

(** Axiom of infinity *)
Parameter infinity_set : S.

Definition infinity_property s := 
  elem empty_set s /\
  forall x, 
    elem x s ->
    elem (union2_set x (single_set x)) s.

Axiom infinity : infinity_property infinity_set.

Definition naturals := 
  sub_set 
    (
      fun s => 
        s = empty_set \/ 
        exists s', 
          s = union2_set s' (single_set s')
    ) 
    infinity_set.

(** Axiom of choice *)
Axiom choice : 
  forall a, 
    ~ elem empty_set a -> 
    (
      forall x y z, 
        elem x a -> 
        elem y a ->
        elem z x -> 
        elem z y -> 
        x = y
    )
    ->
    {b : S | 
      forall x, 
        elem x a -> 
        exists y, 
          elem y x /\ 
          elem y b /\
          forall z, 
            elem z x ->
            elem z b ->
            z = y
    }.
